const assert = require('chai').assert;
const expect = require('chai').expect;
const GamePlay = require("../../lib/k2poker/game_play.js").GamePlay;
const Deck = require("../../lib/k2poker/deck.js").Deck;
const Hand = require("pokersolver").Hand;
const starting_deck = new Deck().cards;

describe('GamePlay', function() {

  describe("#initial state", function () {
    let k2_poker = new GamePlay({"player_1": {"id": 111}, "player_2": {"id": 222}});

    it("should have 2 players", function() {
      assert.equal(k2_poker.players.length, 2);
    });

    it("should have first player (players[0])", function() {
      assert.equal(k2_poker.players[0].id, 111);
      assert.equal(k2_poker.players[0].cards.length, 0)
    });

    it("should have seconde player (players[1])", function() {
      assert.equal(k2_poker.players[1].id, 222);
      assert.equal(k2_poker.players[1].cards.length, 0)
    });

    it("should have a deck of shuffled cards", function() {
      assert.equal(k2_poker.deck.length, 52);
      expect(k2_poker.deck.join()).to.not.equal(starting_deck.join());
    });

    it("should have a game_status of start", function() {
      assert.equal(k2_poker.game_status, "start");
    });

    it("should not have any table cards yet", function() {
      assert.equal(k2_poker.table_cards.length, 0);
    });

  });

  describe("#deal", function () {
    let player_1 = {"id": 111};
    let player_2 = {"id": 222};
    let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
    k2_poker.deal();

    it("should deal 2 cards to player 1 -> " + k2_poker.players[0].cards.join(" "), function() {
      assert.equal(k2_poker.players[0].cards.length, 2)
    });

    it("should deal 2 cards to player 2 -> " + k2_poker.players[1].cards.join(" "), function() {
      assert.equal(k2_poker.players[1].cards.length, 2)
    });

    it("should set the game_status to deal", function(){
      assert.equal(k2_poker.game_status, "deal");
    });

    it("should have 48 cards left in the deck", function(){
      assert.equal(k2_poker.deck.length, 48);
    });

  });

  describe("#discard", function(){

    it("should not allow the player to discard on status 'start'", function(){
      let player_1 = {"id": 111}
      let player_2 = {"id": 222}
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      let old_cards = k2_poker.players[0].cards.slice(0);
      k2_poker.discard(player_1["id"], 0);
      expect(old_cards[0]).to.equal(k2_poker.players[0].cards[0]);
    });

    it("should allow the player to discard card 1 on status 'deal'", function(){
      let player_1 = {"id": 111}
      let player_2 = {"id": 222}
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      k2_poker.deal();
      let old_cards = k2_poker.players[0].cards.slice(0);
      k2_poker.discard(player_1["id"], 0);
      expect(old_cards[0]).to.not.equal(k2_poker.players[0].cards[0]);
      expect(old_cards[1]).to.equal(k2_poker.players[0].cards[1]);
    });

    it("should allow the player to discard card 2 on status 'deal'", function(){
      let player_1 = {"id": 111}
      let player_2 = {"id": 222}
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      k2_poker.deal();
      let old_cards = k2_poker.players[0].cards.slice(0);
      k2_poker.discard(player_1["id"], 1);
      expect(old_cards[0]).to.equal(k2_poker.players[0].cards[0]);
      expect(old_cards[1]).to.not.equal(k2_poker.players[0].cards[1]);
    });

    it("should allow the player to discard card 2 on status 'flop'", function(){
      let player_1 = {"id": 111}
      let player_2 = {"id": 222}
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      k2_poker.deal();
      k2_poker.flop();
      let old_cards = k2_poker.players[0].cards.slice(0);
      k2_poker.discard(player_1["id"], 1);
      expect(old_cards[0]).to.equal(k2_poker.players[0].cards[0]);
      expect(old_cards[1]).to.not.equal(k2_poker.players[0].cards[1]);
    });

    it("should allow the player to discard any card on status 'turn'", function(){
      let player_1 = {"id": 111}
      let player_2 = {"id": 222}
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      k2_poker.deal();
      k2_poker.flop();
      k2_poker.turn();
      let old_cards = k2_poker.players[0].cards.slice(0);
      k2_poker.discard(player_1["id"], 1);
      expect(old_cards[0]).to.equal(k2_poker.players[0].cards[0]);
      expect(old_cards[1]).to.not.equal(k2_poker.players[0].cards[1]);
    });

    //TODO maybe it should allow, but just discard 2?
    it("should not allow the player to discard on the 'river'", function(){
      let player_1 = {"id": 111}
      let player_2 = {"id": 222}
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      k2_poker.deal();
      k2_poker.flop();
      k2_poker.turn();
      k2_poker.river();
      let old_cards = k2_poker.players[0].cards.slice(0);
      assert.equal(k2_poker.discard(player_1["id"], 1), false);
      expect(old_cards[0]).to.equal(k2_poker.players[0].cards[0]);
    });

    it("should dis-allow the player to discard after already discarding", function(){
      let player_1 = {"id": 111}
      let player_2 = {"id": 222}
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      k2_poker.deal();
      k2_poker.discard(player_1["id"], 1);
      let old_cards = k2_poker.players[0].cards.slice(0);

      assert.equal(k2_poker.discard(player_1["id"], 0), false)
      expect(old_cards[0]).to.equal(k2_poker.players[0].cards[0]);
    });

    it("should allow the player to discard/burn at every (allowed) stage", function(){
      let player_1 = {"id": 111}
      let player_2 = {"id": 222}
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      k2_poker.deal();
      expect(k2_poker.discard(player_1["id"], 1)).to.not.equal(false);
      //including after waiting
      k2_poker.player_waiting(player_1["id"]);
      k2_poker.flop();
      expect(k2_poker.discard(player_1["id"], 1)).to.not.equal(false);
      k2_poker.turn();
      expect(k2_poker.discard(player_1["id"], 1)).to.not.equal(false);
      k2_poker.river();
      expect(k2_poker.burn(player_1["id"])).to.not.equal(false);
    });

    it("should allow the player to burn on the 'river'", function(){
      let player_1 = {"id": 111}
      let player_2 = {"id": 222}
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      k2_poker.deal();
      k2_poker.flop();
      k2_poker.turn();
      k2_poker.river();
      let old_cards = k2_poker.players[0].cards.slice(0);
      k2_poker.burn(player_1["id"]);
      expect(old_cards[0]).to.not.equal(k2_poker.players[0].cards[0]);
      expect(old_cards[1]).to.not.equal(k2_poker.players[0].cards[1]);
    });

    // once a player has commited move, set status to waiting
    it("should set the players status to waiting - pending", function(){
      let player_1 = {"id": 111}
      let player_2 = {"id": 222}
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      k2_poker.deal();
      k2_poker.player_waiting(player_1["id"]);
      expect(k2_poker.players[0].status).to.equal("waiting");
    });

  });

  describe("#flop", function () {
    let deck = new Deck().shuffle();
    let player_1 = {"id": 111, "cards": [deck.shift(), deck.shift()]}
    let player_2 = {"id": 222, "cards": [deck.shift(), deck.shift()]}
    let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2, "game_status": "deal", "deck": deck});
    k2_poker.flop();

    it("player 1 should have 2 cards -> " + k2_poker.players[0].cards.join(" "), function() {
      assert.equal(k2_poker.players[0].cards.length, 2)
    });

    it("player 2 should have 2 cards -> " + k2_poker.players[1].cards.join(" "), function() {
      assert.equal(k2_poker.players[1].cards.length, 2)
    });

    it("the table should have 3 cards - The Flop ("+ k2_poker.table_cards.join(" ") +")", function() {
      assert.equal(k2_poker.table_cards.length, 3);
    });

    it("should set the game_status to flop", function(){
      assert.equal(k2_poker.game_status, "flop");
    });

    it("should have 44 cards left in the deck", function(){
      assert.equal(k2_poker.deck.length, 44);
    });

  });

  describe("#turn", function () {
    let player_1 = {"id": 111}
    let player_2 = {"id": 222}
    let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
    k2_poker.deal();
    k2_poker.flop();
    k2_poker.turn();

    it("player 1 should have 2 cards -> " + k2_poker.players[0].cards.join(" "), function() {
      assert.equal(k2_poker.players[0].cards.length, 2)
    });

    it("player 2 should have 2 cards -> " + k2_poker.players[1].cards.join(" "), function() {
      assert.equal(k2_poker.players[1].cards.length, 2)
    });

    it("the table should have 4 cards - The Turn ("+ k2_poker.table_cards.join(" ") +")", function() {
      assert.equal(k2_poker.table_cards.length, 4);
    });

    it("should set the game_status to turn", function(){
      assert.equal(k2_poker.game_status, "turn");
    });

    it("should have 42 cards left in the deck", function(){
      assert.equal(k2_poker.deck.length, 42);
    });

  });

  describe("#river", function () {
    let player_1 = {"id": 111}
    let player_2 = {"id": 222}
    let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
    k2_poker.deal();
    k2_poker.flop();
    k2_poker.turn();
    k2_poker.river();

    it("player 1 should have 2 cards -> " + k2_poker.players[0].cards.join(" "), function() {
      assert.equal(k2_poker.players[0].cards.length, 2)
    });

    it("player 2 should have 2 cards -> " + k2_poker.players[1].cards.join(" "), function() {
      assert.equal(k2_poker.players[1].cards.length, 2)
    });

    it("the table should have 5 cards - The River ("+ k2_poker.table_cards.join(" ") +")", function() {
      assert.equal(k2_poker.table_cards.length, 5);
    });

    it("should set the game_status to river", function(){
      assert.equal(k2_poker.game_status, "river");
    });

    it("should have 40 cards left in the deck", function(){
      assert.equal(k2_poker.deck.length, 40);
    });

  });

  describe("#end", function () {
    let player_1 = {"id": 111}
    let player_2 = {"id": 222}
    let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
    k2_poker.deal();
    k2_poker.flop();
    k2_poker.turn();
    k2_poker.river();
    k2_poker.end();

    it("player 1 should have 2 cards -> " + k2_poker.players[0].cards.join(" "), function() {
      assert.equal(k2_poker.players[0].cards.length, 2)
    });

    it("player 2 should have 2 cards -> " + k2_poker.players[1].cards.join(" "), function() {
      assert.equal(k2_poker.players[1].cards.length, 2)
    });

    it("the table should have 5 cards - End Game ("+ k2_poker.table_cards.join(" ") +")", function() {
      assert.equal(k2_poker.table_cards.length, 5);
    });

    it("should set the game_status to river", function(){
      assert.equal(k2_poker.game_status, "end");
    });

    it("should have 40 cards left in the deck", function(){
      assert.equal(k2_poker.deck.length, 40);
    });

    it("should set winners array to the winner", function(){
      let player_1_hand = Hand.solve(k2_poker.players[0].cards.concat(k2_poker.table_cards));
      let player_2_hand = Hand.solve(k2_poker.players[1].cards.concat(k2_poker.table_cards));
      let winners = Hand.winners([player_1_hand, player_2_hand]);
      if(k2_poker.winners.length == 1) {
        assert.equal(k2_poker.winners[0]["cards"].length, 5);
        assert.equal(k2_poker.winners[0]["hand_type"], winners[0].name);
        assert.equal(k2_poker.winners[0]["description"], winners[0].descr);
        assert.equal(k2_poker.winners[0]["rank"], winners[0].rank);
        if(winners[0] == player_1_hand){
          assert.equal(k2_poker.winners[0]["id"], k2_poker.players[0]["id"]);
          assert.equal(k2_poker.winners[0]["winner"], 1);
        } else {
          assert.equal(k2_poker.winners[0]["id"], k2_poker.players[1]["id"]);
          assert.equal(k2_poker.winners[0]["winner"], 2);
        }
      } else {
        assert.equal(k2_poker.winners[0]["winner"], "draw");
      }

    });

  });

  describe("#to_json should return a representation of the current game", function(){

    it("should return a new game", function(){
      let player_1 = {"id": 111};
      let player_2 = {"id": 222};
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      k2_poker.deal();
      k2_poker.flop();
      json = k2_poker.to_json()
      expect(json["game_status"]).to.equal("flop");
      expect(json["deck"].length).to.equal(44)
      expect(json["table_cards"].length).to.equal(3);
      expect(json["player_1"]["id"]).to.equal(111);
      expect(json["player_1"]["status"]).to.equal("ready");
      expect(json["player_1"]["cards"].length).to.equal(2);
      expect(json["player_2"]["id"]).to.equal(222);
      expect(json["player_2"]["status"]).to.equal("ready");
      expect(json["player_2"]["cards"].length).to.equal(2);
      expect(json["winners"].length).to.equal(0);
    });

    it("should be in a format that can used to pass back to the initializer to continue a new game", function(){
      let player_1 = {"id": 111};
      let player_2 = {"id": 222};
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2});
      k2_poker.deal();
      k2_poker.flop();
      json = k2_poker.to_json();
      k2_reloaded = new GamePlay(json)
      expect(k2_reloaded.game_status).to.equal("flop");
      expect(k2_reloaded.players[0]["id"]).to.equal(111);
      expect(k2_reloaded.players[0]["cards"]).to.equal(k2_poker.players[0]["cards"]);
      expect(k2_reloaded.players[1]["cards"]).to.equal(k2_poker.players[1]["cards"]);
      expect(k2_reloaded.players[1]["status"]).to.equal("ready");
      expect(k2_reloaded.table_cards.length).to.equal(3);
      expect(k2_reloaded.deck.length).to.equal(44);
    });

  });

  describe("winners draw", function(){

    it("should set both users to winners array when its a draw", function(){
      let table_cards = ["5c", "5h", "Qs", "As", "3d"];
      let player_1 = {"id": 111, cards: ["5d", "Qh"], status: "waiting"};
      let player_2 = {"id": 222, cards: ["5s", "Qc"], status: "waiting"};
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2, "table_cards": table_cards, "game_status": "river"});
      json = k2_poker.end().to_json();
      console.log(json);
      expect(json["winners"].length).to.eql(2);
      expect(json["winners"][0]["winner"]).to.eql("draw");
      expect(json["winners"][1]["winner"]).to.eql("draw");
    });

  });

  describe("#parse_hand (fixes return card issues from PokerSolver winning hand and returns an array)", function(){
    it("should return an array of 5 cards (pokersolver returns 7 cards on a straight flush)", function(){
      let player_1 = {"id": 111, cards: ["Ts", "Js"], status: "waiting"};
      let player_2 = {"id": 222, cards: ["Td", "Jd"], status: "waiting"};
      let table_cards = ["As", "Ks", "Qs", "2d", "3d"];
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2, "table_cards": table_cards, "game_status": "river"});
      json = k2_poker.end().to_json();
      expect(json["winners"][0]["cards"]).to.eql(["As","Ks","Qs", "Js", "Ts"]);
    });

    it("should return card value 10 as T (pokersolver returns 10s when passed Ts)", function(){
      let player_1 = {"id": 111, cards: ["Ts", "Js"], status: "waiting"};
      let player_2 = {"id": 222, cards: ["4d", "Jd"], status: "waiting"};
      let table_cards = ["Th", "Tc", "Td", "2d", "3d"];
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2, "table_cards": table_cards, "game_status": "river"});
      json = k2_poker.end().to_json();
      expect(json["winners"]["0"]["cards"]).to.eql(["Ts","Th","Tc","Td","Js"]);
    });

  });

  describe("#hand_type (fixes name of hand_type to 'Royal Flush', which pokersolver returns as 'Straight Flush'", function(){

    it("should be called 'Royal Flush'", function(){
      let player_1 = {"id": 111, cards: ["Ts", "Js"], status: "waiting"};
      let player_2 = {"id": 222, cards: ["Td", "Jd"], status: "waiting"};
      let table_cards = ["As", "Ks", "Qs", "2d", "3d"];
      let k2_poker = new GamePlay({"player_1": player_1, "player_2": player_2, "table_cards": table_cards, "game_status": "river"});
      json = k2_poker.end().to_json();
      expect(json["winners"]["0"]["cards"]).to.eql(["As","Ks","Qs", "Js", "Ts"]);
      expect(json["winners"]["0"]["hand_type"]).to.equal("Royal Flush");
    });

  });

});
